//*********************************************************************************
// The routines in this file are used to process single lines of C code.
// The main routine is getcline, which extracts a single C line.  It determines
// the number of nested levels and the indentatin that should be used.  It
// also determines the type of C line that is being returned.
// These routine handle limit use of #include and conditional compiling.
//*********************************************************************************

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "links.h"
#include "cline.h"
#include "strsubs.h"
#include "tokens.h"
#include "vars.h"
#include "cdirect.h"

#define TERM_UNKNOWN             0
#define TERM_SEMI                1
#define TERM_PAREN               2
#define TERM_NEWLINE             3
#define TERM_ELSE                4
#define TERM_BRACE               5
#define TERM_NEWLINE_OR_COMMENT  6
#define TERM_COLON               7
#define MAX_LEVELS              20

#define INDENT_NONE         0
#define INDENT_SECTION      1 // function, for, while, do, switch or if section
#define INDENT_NEXT_LINE    2 // line after function, for, while, do, switch or if
#define INDENT_CASE_SECTION 3 // case section
#define INDENT_GROUP        4 // stand-along {} group

static char *controllist[] = {"if", "else", 0 };
static char *looplist[] = {"while", "for", "switch", 0 };
static char *typelist[] = {"void", "FILE", "static", "extern", "unsigned", "signed", 0};
LinkT *typeslist = 0;
static int inline_comment = 0;
extern int skipflag;
char objname[32];

static char inbuf[CLINE_MAX_SIZE];
//static char *inptr;
static int endoffile;
static int level;
static int spinlevel;
static char indent[MAX_LEVELS];
static char looplevel[MAX_LEVELS];
static int initflag = 0;
static int clinetype;
static LinkT *endptr[MAX_LEVELS];
static int cstate = CSTATE_IDLE;
static int prevtype = 0;
static int prevstate = 0;

static LinkT *intokens = 0;

extern FILE *inputfile[10];
extern int inputline[10];
extern char inputfname[10][80];
extern int inputfileindex;
extern ConInfoT *conlist;

void PrintFileLine(void)
{
    printf("%s: Line %d", inputfname[inputfileindex], inputline[inputfileindex]);
}

static LinkT *GetLine(void)
{
    LinkT *tokenlist = 0;

    while (tokenlist == 0)
    {
        if (fgets(inbuf, 1000, inputfile[inputfileindex]) == 0)
        {
            fclose(inputfile[inputfileindex]);
	    if (inputfileindex == 0)
	    {
		//if (CheckForMoreFiles()) continue;
                endoffile = 1;
                break;
	    }
	    inputfileindex--;
	    continue;
        }
	inputline[inputfileindex]++;
        RemoveCRLF(inbuf);
	//printf("Line %d: %s\n", inputline[inputfileindex], inbuf);
        getctokens(inbuf, cstate == CSTATE_C_COMMENTS, &tokenlist);
    }
    return tokenlist;
}

int BalancedParens(void)
{
    int parencount = 0;
    LinkT *link = intokens;

    while (link)
    {
	if (link->str[0] == '(') parencount++;
	else if (link->str[0] == ')') parencount--;
	link = link->next;
    }
    return (parencount <= 0);
}

static void CheckLine(void)
{
    LinkT *tokenlist;

    if (intokens) return;

    intokens = GetLine();

    if (endoffile || !intokens) return;

    // Perform macro expandsion on input line
    if (intokens->str[0] != '#' && !skipflag)
    {
	while (!BalancedParens())
	{
	    tokenlist = GetLine();
            if (endoffile || !tokenlist) return;
	    ListAppend(intokens, tokenlist);
	}

        intokens = ExpandLine(intokens);
    }
}

static void CheckForInlineComment(int clinetype, LinkT *tokenlist)
{
    char *inptr;

    if (intokens == 0)
    {
	inline_comment = 0;
	return;
    }
    inptr = intokens->str;
    if ((inptr[0] == '/' && inptr[1] == '/') ||
        (inptr[0] == '/' && inptr[1] == '*'))
    {
	inline_comment = 1;
    }
    else
    {
        inline_comment = 0;
	return;
    }

    // Clear inline_comment if a directive.  We should just gobble it.
    // The only valid case would be a numberic define.
    if (clinetype != CLINE_DIRECTIVE) return;
    if (tokenlist == 0 || tokenlist->next == 0) return;
    tokenlist = tokenlist->next;

    if (strcmp(tokenlist->str, "define") != 0) inline_comment = 0;
}

static LinkT *MoveToken(LinkT *tokenlist)
{
    LinkT *next;

    if (intokens == 0)
    {
	printf("MoveToken: tokenlist is NULL\n");
	exit(0);
    }

    //printf("MoveToken: %s\n", intokens->str);

    next = intokens->next;
    if (tokenlist == 0)
    {
	tokenlist = intokens;
    }
    else
    {
	LinkT *tail = ListTail(tokenlist);
	tail->next = intokens;
    }
    intokens->next = 0;
    intokens = next;

#if 0
    next = tokenlist;
    while(next)
    {
	printf(" -%s-", next->str);
	next = next->next;
	if (++i > 20)break;
    }

    printf(" --");
    next = intokens;

    while(next)
    {
	printf(" -%s-", next->str);
	next = next->next;
	if (++i > 20)break;
    }

    printf("\n");
#endif

    return tokenlist;
}

static LinkT *MoveTokenCheckLine(LinkT *tokenlist)
{
    tokenlist = MoveToken(tokenlist);
    CheckLine();
    return tokenlist;
}

static void GobbleComment(void)
{
    int len;
    char *inptr;

    if (intokens == 0) return;

    inptr = intokens->str;

    if (inptr[0] != '/') return;

    if (inptr[1] == '/')
    {
	intokens = intokens->next;
	return;
    }

    if (inptr[1] != '*') return;

    while (1)
    {
	CheckLine();
	if (endoffile) break;
	if (intokens == 0) break;
	inptr = intokens->str;
        len = strlen(inptr);
	intokens = intokens->next;
	if (inptr[len-2] == '*' && inptr[len-1] == '/') break;
    }
}

static int CopyToTerminator(LinkT **ptokenlist, int term, int declflag)
{
    int val;
    char *inptr;
    int parencount = 0;
    int gobblecomments = (term == TERM_SEMI) ||
        (term == TERM_PAREN) || (term == TERM_COLON) || declflag;


    if (intokens == 0) return term;

    //printf("CopyToTerminator: %d %s\n", term, inptr);
    //printf("CopyToTerminator 1\n");

    if (clinetype == CLINE_TYPEDEF)
    {
	*ptokenlist = MoveTokenCheckLine(*ptokenlist);
        if (endoffile) return term;
    }

    inptr = intokens->str;

    if (clinetype == CLINE_C_COMMENTS)
    {
        int len = strlen(inptr);
        //printf("Check if end of CLINE_C_COMMENTS\n");
        if (inptr[len-2] == '*' && inptr[len-1] == '/')
        {
	    clinetype = CLINE_C_COMMENT;
	    //printf("Yessssssss\n");
        }
	//else printf("Nooooooo\n");
	*ptokenlist = MoveToken(*ptokenlist);
	return term;
    }

    if (strcmp(inptr, "struct") == 0)
    {
	*ptokenlist = MoveTokenCheckLine(*ptokenlist);
        if (endoffile) return term;

        inptr = intokens->str;

	if (ListFindLinkString(typeslist, inptr) == 0)
	    term = TERM_BRACE;
	else
	    term = TERM_SEMI;
	if (clinetype == CLINE_TYPEDEF)
            clinetype = CLINE_TYPE_STRUCT;
	else if (term == TERM_BRACE)
            clinetype = CLINE_STRUCT;
	else
	    clinetype = CLINE_DECL_STRUCT;
    }

    while (1)
    {
    //printf("CopyToTerminator 2\n");
        CheckLine();
        if (endoffile) break;
	inptr = intokens->str;
    //printf("CopyToTerminator 3\n");

	if (gobblecomments && inptr[0] == '/' &&
	    (inptr[1] == '/' || inptr[1] == '*'))
	{
	    GobbleComment();
            if (endoffile) break;
	    continue;
	}
	inptr = intokens->str;

	if (inptr[0] == '\\' && inptr[1] == 0)
	{
	    intokens = intokens->next;
	    continue;
	}

    //printf("CopyToTerminator 3a\n");
        val = *inptr++;
        *ptokenlist = MoveToken(*ptokenlist);

    //printf("CopyToTerminator 3b\n");
        if (intokens == 0)
	{
	    if (term == TERM_NEWLINE || term == TERM_NEWLINE_OR_COMMENT)
	        break;
	}
	else if (term == TERM_NEWLINE_OR_COMMENT)
	{
	    if (intokens->str[0] == '/' &&
	        (intokens->str[1] == '/' || intokens->str[1] == '*'))
	    {
		break;
	    }
	}

    //printf("CopyToTerminator 4\n");

	if (term == TERM_BRACE)
	{
	    CheckLine();
            if (endoffile) break;
	    if (intokens->str[0] == '{') break;
        }
        else if (term == TERM_UNKNOWN)
        {
            if (val == '(')
            {
                parencount++;
                if (declflag)
                    term = TERM_PAREN;
                else
                    term = TERM_SEMI;
		gobblecomments = 1;
            }
            else if (val == ',' || val == '=')
            {
	        term = TERM_SEMI;
		gobblecomments = 1;
            }
            else if (val == ';')
            {
                term = TERM_SEMI;
                break;
            }
            else if (val == '{') break;
            else if (val == '}') break;
        }
        else if (term == TERM_PAREN)
        {
            if (val == '(') parencount++;
            else if (val == ')')
            {
                parencount--;
                if (parencount < 1) break;
            }
        }
        else if (term == TERM_SEMI)
        {
            if (val == ';') break;
        }
        else if (term == TERM_COLON)
        {
            if (val == ':') break;
        }
        else if (term == TERM_ELSE)
        {
	    CheckLine();
            if (endoffile) break;
	    if (strcmp(intokens->str, "if") != 0) break;
            term = TERM_PAREN;
        }
    //printf("CopyToTerminator 5\n");
    }

    //printf("CopyToTerminator 6\n");
    return term;
}

static LinkT *GetUnit(int *pnewline)
{
    int declflag = 0;
    int printflag = 0;
    char *inptr;
    LinkT *tokenlist = 0;
    int term = TERM_UNKNOWN;
    //static char buffer[CLINE_MAX_SIZE];

    *pnewline = 1;

    CheckLine();
    if (endoffile) return 0;
    inptr = intokens->str;

    //printf("GetUnit 1\n");

    clinetype = CLINE_UNKNOWN;

    if (cstate == CSTATE_C_COMMENTS)
    {
	clinetype = CLINE_C_COMMENTS;
	term = TERM_NEWLINE;
    }
    else if (skipflag == SKIP_MODE)
    {
	clinetype = CLINE_SKIP;
        term = TERM_NEWLINE;
    }
    else if (inptr[0] == '#')
    {
	clinetype = CLINE_DIRECTIVE;
	term = TERM_NEWLINE_OR_COMMENT;
    }
    else if (inptr[0] == '/' && inptr[1] == '*')
    {
	clinetype = CLINE_C_COMMENTS;
	term = TERM_NEWLINE;
    }
    else if (inptr[0] == '/' && inptr[1] == '/')
    {
        if (!strncmp(inptr, "// SPIN OBJECT", 14))
        {
            char *ptr1 = SkipChar(&inptr[14], ' ');
            char *ptr2 = FindChar(ptr1, ' ');
            int len = (int)ptr2 - (int)ptr1;
            memcpy(objname, ptr1, len);
            objname[len] = 0;
            //printf("objname = %s\n", objname);
            AddToSymbolList(objname);
        }
        clinetype = CLINE_COMMENT;
        term = TERM_NEWLINE;
    }
    else if (SearchListDelimiter(inptr, controllist, " ({\t") >= 0)
    {
        clinetype = CLINE_CONTROL;
        if (strcmp(inptr, "else") == 0 && (
            inptr[4] == 0 || inptr[4] == ' ' || inptr[4] == '\t'))
            term = TERM_ELSE;
        else
            term = TERM_PAREN;
    }
    else if (SearchListDelimiter(inptr, looplist, " ({\t") >= 0)
    {
        clinetype = CLINE_LOOP;
        term = TERM_PAREN;
    }
    else if ((strcmp(inptr, "case") == 0) ||
        (strcmp(inptr, "default") == 0))
    {
        clinetype = CLINE_CASE;
        term = TERM_COLON;
    }
    else if (SearchList(inptr, typelist, SEARCH_EXACT) >= 0 ||
	ListFindLinkString(typeslist, inptr))
    {
        declflag = 1;
    }
    else if (strcmp(inptr, "typedef") == 0)
    {
	clinetype = CLINE_TYPEDEF;
	term = TERM_SEMI;
    }
    //printf("GetUnit 2\n");

    if (printflag) printf("YYY %d\n", clinetype);

    term = CopyToTerminator(&tokenlist, term, declflag);
    if (tokenlist == 0) return 0;

    //printf("GetUnit 3\n");
    if (inline_comment) *pnewline = 0;

#if 0
    printf("tokenlist->str = %s\n", tokenlist->str);
    if (intokens == 0)
        printf("intokens = 0\n");
    else
        printf("intokens->str = %s\n", intokens->str);
#endif
    CheckForInlineComment(clinetype, tokenlist);
    //printf("GetUnit 4\n");

    if (tokenlist->str[0] == '{') clinetype = CLINE_BRACE_LEFT;
    else if (tokenlist->str[0] == '}') clinetype = CLINE_BRACE_RIGHT;
    else if (declflag)
    {
        if(term == TERM_PAREN)
        {
            // Check if function prototype
	    CheckLine();
	    if (endoffile) return 0;

            if (intokens->str[0] == ';')
            {
                tokenlist = MoveToken(tokenlist);
                clinetype = CLINE_PROTOTYPE;
            }
            else
            {
                clinetype = CLINE_FUNCTION;
            }
        }
        else if (level == 0) clinetype = CLINE_DECL_GLOBAL;
        else clinetype = CLINE_DECL_LOCAL;
    }
    //printf("GetUnit 5\n");

    if (printflag) printf("YYY %d\n", clinetype);

    if (clinetype == CLINE_UNKNOWN && tokenlist != 0)
    {
	LinkT *tail = ListTail(tokenlist);
	if (tail->str[0] == ';') clinetype = CLINE_STATEMENT;
    }

    if (printflag) printf("YYY %d\n", clinetype);

#if 0
    {
	LinkT *link = tokenlist;

	printf("GetUnit:");
	while (link)
	{
	    printf(" \"%s\"", link->str);
	    link = link->next;
	}
	printf("\n");
    }
#endif

    //printf("GetUnit 6\n");
    //getctokens(buffer, clinetype, &tokenlist);
    return tokenlist;
}

static int GetIndentLevel(int clinetype)
{
    int level0 = level;
    int spinlevel0 = spinlevel;

    if (clinetype == CLINE_CASE &&
        indent[level0] == INDENT_CASE_SECTION) spinlevel0--;
    else if (clinetype == CLINE_BRACE_LEFT &&
        indent[level0] != INDENT_SECTION && indent[level0] != INDENT_GROUP )
	spinlevel0--;
    else if (clinetype == CLINE_BRACE_RIGHT)
    {
        while (indent[level0] != INDENT_SECTION &&
	    indent[level0] != INDENT_GROUP)
	{
	    level0--;
	    spinlevel0--;
	}
        if (level0 > 0)
	{
	    if (indent[level0] != INDENT_GROUP) spinlevel0--;
	    level0--;
	}
    }
    return spinlevel0;
}

static void UpdateIndentLevel(int clinetype, int *oneliner)
{
    int bumpindent = INDENT_NONE;

    *oneliner = 0;

    if (clinetype == CLINE_FUNCTION || clinetype == CLINE_CONTROL ||
        clinetype == CLINE_LOOP)
	bumpindent = INDENT_NEXT_LINE;
    else if (clinetype == CLINE_CASE && indent[level] != INDENT_CASE_SECTION)
        bumpindent = INDENT_CASE_SECTION;

    if (clinetype == CLINE_BRACE_LEFT)
    {
#if 0
        if (indent[level] == INDENT_SECTION)
	{
	    level++;
	    endptr[level] = endptr[level-1];
	    looplevel[level] = 0;
	}
        indent[level] = INDENT_SECTION;
#else
        if (indent[level] == INDENT_NEXT_LINE)
	{
            indent[level] = INDENT_SECTION;
	}
	else
	{
	    level++;
	    endptr[level] = endptr[level-1];
	    looplevel[level] = 0;
            indent[level] = INDENT_GROUP;
	}
#endif
    }
    else if (clinetype == CLINE_BRACE_RIGHT)
    {
        while (indent[level] != INDENT_SECTION &&
	    indent[level] != INDENT_GROUP)
	{
	    level--;
	    spinlevel--;
	}
        if (level > 0)
	{
	    if (indent[level] != INDENT_GROUP) spinlevel--;
	    level--;
	}
    }
    else if (bumpindent != INDENT_NONE)
    {
	spinlevel++;
        indent[++level] = bumpindent;
	if (clinetype == CLINE_LOOP)
	{
            looplevel[level] = 1;
	    endptr[level] = 0;
	}
	else
	{
	    looplevel[level] = 0;
	    endptr[level] = endptr[level-1];
	}
    }
    else if (indent[level] == INDENT_NEXT_LINE)
    {
	*oneliner = 1;
	level--;
	spinlevel--;
    }
}

int UpdateAndGetIndentLevel(int clinetype, int *oneliner)
{
    int retval = GetIndentLevel(clinetype);
    UpdateIndentLevel(clinetype, oneliner);
    return retval;
}

void initcline(void)
{
    int i;

    inline_comment = 0;
    //skipflag = SKIP_NONE;
    clinetype = 0;
    cstate = CSTATE_IDLE;
    prevtype = 0;
    prevstate = 0;
    //dlevel = 0;
    level = 0;
    spinlevel = 0;
    indent[0] = INDENT_SECTION;
    looplevel[0] = -1;
    for (i = 0; i < MAX_LEVELS; i++)
        endptr[i] = 0;
    endoffile = 0;
    //inptr = inbuf;
    intokens = 0;
    //*inptr = 0;
    initflag = 1;
    //typeslist = 0;
    //EmptyLink = LinkCreate("", sizeof(LinkT));
    //AddToConList("CSPIN_FLAG", EmptyLink);
    objname[0] = 0;
}

void PrintCLineCState(char *buffer)
{
#include "clinename.h"

    printf("getcline: %s - %s, %s\n",
        buffer, clinename[clinetype], cstatename[cstate]);
}

void UpdateCState(void)
{
    // Handle comment continuation lines
    if (clinetype == CLINE_C_COMMENTS)
    {
        cstate = CSTATE_C_COMMENTS;
	return;
    }
    else if (cstate == CSTATE_C_COMMENTS)
    {
	cstate = prevstate;
    }

    // Don't update the cstate for other comments
    if (clinetype == CLINE_COMMENT || clinetype == CLINE_C_COMMENT) return;

    // Process non-comment lines
    if (clinetype == CLINE_FUNCTION)
        cstate = CSTATE_FUNCTION;
    else if (clinetype == CLINE_STRUCT)
        cstate = CSTATE_STRUCT;
    else if (clinetype == CLINE_TYPE_STRUCT)
        cstate = CSTATE_TYPE_STRUCT;
    else if (cstate == CSTATE_FUNCTION)
    {
	if (level == 0 && clinetype != CLINE_BRACE_RIGHT)
	    cstate = CSTATE_IDLE;
    }
    else if (cstate == CSTATE_STRUCT)
    {
	if (prevtype == CLINE_BRACE_RIGHT && clinetype == CLINE_STATEMENT)
	{
	    clinetype = CLINE_STRUCT_VAR;
	    cstate = CSTATE_IDLE;
	}
	else if (clinetype == CLINE_DECL_LOCAL ||
	    clinetype == CLINE_DECL_GLOBAL || clinetype == CLINE_DECL_STRUCT)
	    clinetype = CLINE_STRUCT_ELEM;
    }
    else if (cstate == CSTATE_TYPE_STRUCT)
    {
	if (prevtype == CLINE_BRACE_RIGHT && clinetype == CLINE_STATEMENT)
	{
	    clinetype = CLINE_STRUCT_TYPE;
	    cstate = CSTATE_IDLE;
	}
	else if (clinetype == CLINE_DECL_LOCAL ||
	    clinetype == CLINE_DECL_GLOBAL || clinetype == CLINE_DECL_STRUCT)
	    clinetype = CLINE_STRUCT_ELEM;
    }

    prevstate = cstate;
    prevtype = clinetype;
}

static LinkT *getcline1(int *pclinetype, int *pnewline)
{
    LinkT *tokenlist = 0;

    if (!initflag) initcline();

    if (skipflag == SKIP_LINE) skipflag = SKIP_NONE;

    tokenlist = GetUnit(pnewline);
    if (endoffile) return 0;

    if (skipflag || tokenlist->str[0] == '#')
    {
	//printf("skipflag = %d, str = %s\n", skipflag, tokenlist->str);
        *pclinetype = clinetype;
        return tokenlist;
    }

    // kludge for break in switch statement
    if (strcmp(tokenlist->str, "break") == 0 && indent[level] == INDENT_CASE_SECTION)
    {
        tokenlist = GetUnit(pnewline);
        if (endoffile) return 0;
    }

    // Update State Variable
    UpdateCState();

    *pclinetype = clinetype;
    return tokenlist;
}

LinkT *getcline(int *pclinetype, int *pnewline)
{
    LinkT *tokenlist;

    while ((tokenlist = getcline1(pclinetype, pnewline)))
    {
	if (!ProcessDirective(tokenlist)) break;
    }

    return tokenlist;
}

#if 0
static char *GetSymbol(char *buffer)
{
    char *ptr1 = SkipChars(buffer, " \t");
    char *ptr2 = FindChars(ptr1, " \t");
    static char symbol[100];

    if (ptr2 > ptr1) ptr2--;

    strcpyx(symbol, ptr1, ptr2);

    return symbol;
}
#endif

void GetSymbols(char *str, char **symbols, int num)
{
    while (num-- > 0)
    {
	*symbols++ = str = SkipChars(str, " \t");
	if (*str == '(' || *str == '[' || *str == '"')
	{
	    int quotemode = (*str == '"');
	    int parencount = 1;
	    while (*str && parencount > 0)
	    {
		if (quotemode)
		{
		    if (*str == '"')
		    {
			quotemode = 0;
			parencount--;
		    }
		    else if (*str == '\\')
		    {
			str++;
			if (*str == 0) break;
		    }
		}
		else
		{
	            if (*str == '(' || *str == '[' || *str == '"')
		    {
			parencount++;
	                quotemode = (*str == '"');
		    }
	            else if (*str == ')' || *str == ']') parencount--;
		}
	        str++;
	    }
	    if (*str) str++;
	}
	else str = FindChars(str, " \t");
	if (*str) *str++ = 0;
    }
}

void SetEndPtr(LinkT *list)
{
    endptr[level] = list;
    //printf("SetEndPtr: %s\n");
}

LinkT *GetEndPtr(int offset)
{
    int i = level + offset;
    if (i < 0 || i >= MAX_LEVELS) return 0;
    if (offset == 1 && looplevel[i] == 0) return 0;
    return endptr[i];
}

void checkcline(LinkT *tokenlist, int clinetype)
{
    char *token;
    LinkT *list;

    if (clinetype != CLINE_TYPEDEF) return;

    for (list = tokenlist; list; list = list->next)
    {
	token = list->str;
	if (list == tokenlist && strcmp(token, "typedef") == 0) continue;
	if (SearchList(token, typelist, SEARCH_EXACT) >= 0) continue;
        if (ListFindLinkString(typeslist, token)) continue;
	if (strcmp(token, "*") == 0) continue;
	if (strcmp(token, "**") == 0) continue;
	break;
    }

    if (list) list = list->next;

    if (list == 0 || strcmp(list->str, ";") != 0)
    {
	PrintFileLine();
        printf(" - ERROR: bad typedef\n");
        for (list = tokenlist; list; list = list->next)
	    printf(" %s", list->str);
	printf("\n");
	CspinExit(1);
    }
}
