#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include "links.h"
#include "strsubs.h"
#include "vars.h"
#include "spinit.h"
#include "spinstruct.h"
#include "spinvars.h"
#include "cline.h"
#include "tokens.h"
#include "cdirect.h"

int modeflag = MODEFLAG_NULL;
int prevtype = CLINE_UNKNOWN;
extern int nonewline;
int inlineflag = 0;

extern FILE *outfile;
extern ConInfoT *conlist;
extern VarInfoT *varlist;
extern TypeInfoT *typeslist;

static char *cstr[] = {"=",  "!=", ">=", "<=", "&&",  "||", "!", "break", "while", "if", "switch", "%", "~",  0};
static char *sstr[] = {":=", "<>", "=>", "=<", "and", "or", "not", "quit", "repeat while", "if", "case", "//", "!",  0};
static char *operator[] = {"=", "*", "&", "|", "+", "-", "~", "!", "%", "^", "/", "&&", "||", "++", "--", 0};
static char *typelist[] = {"static", "register", "const", "extern", "unsigned", "signed", "struct", 0};
static char *varglist[] = {"printf", "scanf", "fprintf", "fscanf", "sprintf", "sscanf", 0};

void InitSpinit(void)
{
    modeflag = MODEFLAG_NULL;
    prevtype = CLINE_UNKNOWN;
}

int FindNextOperator(int i, char **tokens, int num)
{
    for (; i < num; i++)
    {
	if (SearchList(tokens[i], operator, SEARCH_START) >= 0) break;
    }
    return i;
}

int IsLeftValue(int i, char **tokens, int num)
{
    int len;

    i = FindNextOperator(i, tokens, num);

    if (i >= num) return 0;
    len = strlen(tokens[i]);
    if (tokens[i][len-1] == '=') return 1;
    return 0;
}

int LocateClosingParen(char **tokens, int i, int num1)
{
    int j;
    int parencount = 0;

    if (tokens[i][0] != '(') return i;

    for (j = i;j < num1; j++)
    {
	if (tokens[j][0] == '(') parencount++;
	else if (tokens[j][0] == ')')
	{
	    parencount--;
	    if (parencount <= 0) break;
	}
    }
    if (j < num1) return j;
    return i;
}

int SizeOf(char *str)
{
    int size = 0;
    VarInfoT *varinfo;
    TypeInfoT *typeinfo;

    if ((varinfo = FindVariable(varlist, str)))
    {
	size = SizeOfVarInfo(varinfo);
    }
    else if ((typeinfo = FindType(typeslist, str)))
    {
	if (typeinfo->ptrlevel > 0)
	    size = 4;
	else
            size = typeinfo->size;
    }

    return size;
}

char *GetSpinVariableName(char *str)
{
    VarInfoT *varinfo = FindVariable(varlist, str);
    ConInfoT *coninfo = (ConInfoT *)ListFindLinkString((LinkT *)conlist, str);

    if (varinfo) str = varinfo->str1;
    else if (coninfo) str = coninfo->str1;
    return str;
}

void CheckEndPtr(LinkT **ptokenlistout, int offset)
{
    LinkT *endptr1 = GetEndPtr(offset);
    if (endptr1)
    {
	endptr1 = ListDuplicate(endptr1);
	//printf("CheckEndPtr: %s\n", endptr);
	if (offset == 1) AddToken(" ", ptokenlistout);
	*ptokenlistout = ListAppend(*ptokenlistout, endptr1);
	if (offset == 0) AddToken("\n", ptokenlistout);
        //AddToken("\n", ptokenlistout);
    }
}

void ConvertFunctionToSpin(char **tokens1, int num1, LinkT **ptokenlistout, int clinetype)
{
    int i;
    int first = 0;
    VarInfoT *vartype;

    modeflag = MODEFLAG_PUB;
    AddToken("\n", ptokenlistout);
    AddToken("PUB", ptokenlistout);
    for (i = 0; i < num1; i++)
    {
        if (tokens1[i][0] == '\n') break;
	if (strcmp(tokens1[i], "extern") == 0) continue;
	if (strcmp(tokens1[i], "const") == 0) continue;
	if (strcmp(tokens1[i], "static") == 0) continue;
	if (strcmp(tokens1[i], "signed") == 0) continue;
	if (strcmp(tokens1[i], "unsigned") == 0) continue;
	if (FindType(typeslist, tokens1[i])) continue;
        if (tokens1[i][0] == '*') continue;
        if (tokens1[i][0] == '(')
        {
	    first = 1;
	    continue;
        }
        if (tokens1[i][0] == ')')
        {
	    if (first != 2) continue;
        }
        if (first == 1)
        {
            AddToken("(", ptokenlistout);
	    first = 2;
        }
	if ((vartype = FindVariable(varlist, tokens1[i])))
	    AddToken(vartype->str1, ptokenlistout);
	else
            AddToken(TranslateCStrToSpinStr(tokens1[i], clinetype), ptokenlistout);
    }
}

void ConvertForToSpin(char **tokens1, int num1, LinkT **ptokenlistout, int clinetype)
{
    char *vptr1 = tokens1[2];
    char *vptr2;
    char *vptr3;
    char *iptr;
    char *tptr;
    int i, semi2, semi3;
    LinkT *tokenlistextra = 0;

    // Locate semicolons
    for (i = 0; i < num1; i++)
	if (tokens1[i][0] == ';') break;

    semi2 = i;
    vptr2 = tokens1[i+1];

    for (i++; i < num1; i++)
	if (tokens1[i][0] == ';') break;

    semi3 = i;
    vptr3 = tokens1[i+1];
    iptr  = tokens1[i+2];

    if (vptr3[0] == '+' || vptr3[0] == '-')
    {
	tptr = iptr;
	iptr = vptr3;
	vptr3 = tptr;
    }

    // Check for simple form
    if (0 && strcmp(vptr1, vptr2) == 0 &&
        strcmp(vptr1, vptr3) == 0 &&
        tokens1[3][0] == '=' &&
	(iptr[0] == '+' || iptr[0] == '-') &&
	semi2 < num1 && semi3 < num1)
    {
        AddToken("repeat", ptokenlistout);
        AddToken(GetSpinVariableName(vptr1), ptokenlistout);
        AddToken("from", ptokenlistout);
	for (i = 4; i < semi2; i++)
            AddToken(GetSpinVariableName(tokens1[i]), ptokenlistout);
        AddToken("to", ptokenlistout);
	for (i = semi2 + 3; i < semi3; i++)
            AddToken(GetSpinVariableName(tokens1[i]), ptokenlistout);
        if (strcmp(tokens1[semi2+2], "<") == 0)
            AddToken("- 1", ptokenlistout);
        else if (strcmp(tokens1[semi2+2], ">") == 0)
            AddToken("+ 1", ptokenlistout);
        if (strcmp(iptr, "--") == 0)
            AddToken("srep -1", ptokenlistout);
        else if (strcmp(iptr, "+=") == 0)
        {
            AddToken("srep", ptokenlistout);
	    for (i = semi3 + 3; i < num1; i++)
	    {
		if (tokens1[i][0] == ')') break;
                AddToken(GetSpinVariableName(tokens1[i]), ptokenlistout);
	    }
        }
        else if (strcmp(iptr, "-=") == 0)
        {
            AddToken("srep -", ptokenlistout);
	    for (i = semi3 + 3; i < num1; i++)
	    {
		if (tokens1[i][0] == ')') break;
                AddToken(tokens1[i], ptokenlistout);
	    }
        }
	//if (!nonewline) AddToken("\n", ptokenlistout);
        return;
    }

    // Else handle complex form as seperate statements
    ConvertStatementToSpin(2, tokens1, semi2, ptokenlistout, clinetype);
    AddToken("\n", ptokenlistout);
    AddToken("repeat", ptokenlistout);
    AddToken("while", ptokenlistout);
    AddToken("(", ptokenlistout);
    ConvertStatementToSpin(semi2+1, tokens1, semi3, ptokenlistout, clinetype);
    AddToken(")", ptokenlistout);
    //if (!nonewline) AddToken("\n", ptokenlistout);

    // Generate the extra part
    ConvertStatementToSpin(semi3+1, tokens1, num1 - 1, &tokenlistextra, clinetype);
    SetEndPtr(tokenlistextra);

}

void ConvertControlToSpin(char **tokens1, int num1, LinkT **ptokenlistout, int clinetype)
{
    int i = 0;

    if (num1 >= 2 && strcmp(tokens1[0], "else") == 0 && strcmp(tokens1[1], "if") == 0)
    {
	AddToken("elseif", ptokenlistout);
	i = 2;
    }
    else if (strcmp(tokens1[0], "for") == 0)
    {
	ConvertForToSpin(tokens1, num1, ptokenlistout, clinetype);
	return;
    }
    else
    {
        AddToken(TranslateCStrToSpinStr(tokens1[0], clinetype), ptokenlistout);
        i = 1;
    }
    ConvertStatementToSpin(0, &tokens1[i], num1 - i, ptokenlistout, clinetype);
}

int CountParms(int i, char **tokens1, int num1)
{
    int j;
    int argcount = 0;
    int parencount = 0;

    // Count the number of parameters for printf
    for (j = i; j < num1; j++)
    {
	if (tokens1[j][0] == '(') parencount++;
	else if (tokens1[j][0] == '[') parencount++;
	else if (tokens1[j][0] == ',' && parencount == 1) argcount++;
	else if (tokens1[j][0] == ']') parencount--;
	else if (tokens1[j][0] == ')')
	{
	    parencount--;
	    if (parencount <= 0) break;
	}
    }
    return argcount;
}

int ConvertStatementToSpin(int i, char **tokens1, int num1, LinkT **ptokenlistout, int clinetype)
{
    int prev_var = 0;
    VarInfoT *varlink;
    ConInfoT *coninfo;
    int indirect = 0;
    int j;

#if 0
    printf("ConvertStatementToSpin:");
    for (j = i; j < num1; j++) printf(" \"%s\"", tokens1[j]);
    printf("\n");
#endif

    for (; i < num1; i++)
    {
	varlink = FindVariable(varlist, tokens1[i]);

        // Check for type casting
        if (i < num1 - 3 && tokens1[i][0] == '(' &&
	    (FindType(typeslist, tokens1[i+1]) ||
	     SearchList(tokens1[i+1], typelist, SEARCH_EXACT) >= 0))
	{
	    int parencount = 1;
	    // Skip over type casting for now
	    for (i++; i < num1; i++)
	    {
		if (tokens1[i][0] == '(')
		    parencount++;
		else if (tokens1[i][0] == ')')
		{
		    parencount--;
		    if (parencount <= 0) break;
		}
	    }
	    continue;
	}
        else if ((j = SearchList(tokens1[i], varglist, SEARCH_EXACT)) >= 0)
	{
	    char *ptr;
	    int argcount = CountParms(i, tokens1, num1);

            if (j > 1) argcount--;
	    ptr = str_alloc(25);
	    sprintf(ptr, "c.%s%d", tokens1[i], argcount);
            AddToken(ptr, ptokenlistout);
	    prev_var = 1;
	    indirect = 0;
	    continue;
	}
        else if (i < num1 - 3 && strcmp(tokens1[i], "sizeof") == 0)
	{
	    char *ptr = str_alloc(12);
	    sprintf(ptr, "%d", SizeOf(tokens1[i+2]));
	    AddToken(ptr, ptokenlistout);
	    i += 3;
	    continue;
	}
	else if (tokens1[i][0] == '(' && tokens1[i+1][0] == ')')
	{
	    i++;
	    continue;
	}
	else if (tokens1[i][0] == '*' && prev_var == 0)
	{
	    i++;
            varlink = FindVariable(varlist, tokens1[i]);
	    if (varlink)
	    {
	        if (varlink->ptrlevel < 2 && varlink->typeinfo && varlink->typeinfo->structvars)
	            i = ConvertStructReferenceToSpin(i, tokens1, num1, ptokenlistout, -1);
                else
		{
	            i = ConvertPointerReferenceToSpin(i, tokens1, num1, ptokenlistout, -1);
		}
	        indirect = 0;
		prev_var = 1;
		continue;
	    }
	}
	else if (varlink && varlink->typeinfo && varlink->typeinfo->structvars)
	{
	    i = ConvertStructReferenceToSpin(i, tokens1, num1, ptokenlistout, indirect);
	    prev_var = 1;
	    indirect = 0;
            continue;
	}
	else if ((varlink && i < num1 - 1 && (varlink->array || tokens1[i+1][0] == '[')) ||
	    (tokens1[i][0] == '*' && prev_var == 0))// || varlink->array)
	{
	    //AddToken("?", ptokenlistout);
	    i = ConvertPointerReferenceToSpin(i, tokens1, num1, ptokenlistout, indirect);
	    prev_var = 1;
	    indirect = 0;
            continue;
	}
	else if (varlink && GetUseLink(varlink) == CVARS_GLOBAL && GetArrayLink(varlink) == 1)
	{
            AddToken("@", ptokenlistout);
	}
	else if (strcmp(tokens1[i], "&") == 0 && prev_var == 0)
	{
	    int j;

            AddToken("@", ptokenlistout);
	    // Check for opening paren and remove if not needed
	    if (tokens1[i+1][0] != '(') continue;
	    j = LocateClosingParen(tokens1, i+1, num1);
	    if (tokens1[j][0] == ')')
	    {
		tokens1[i+1] = "";
		tokens1[j] = "";
	    }
	    //indirect = 1;
            continue;
	}
	else if (strcmp(tokens1[i], "continue") == 0)
	{
	    CheckEndPtr(ptokenlistout, 0);
	    AddToken("next", ptokenlistout);
	    indirect = 0;
	    continue;
	}
	else if (varlink)
	{
	    if (GetVarSizeLink(varlink, 0) < 4 &&
	        varlink->use != CVARS_FUNCTION &&
		varlink->use != CVARS_PROTOTYPE)
	    {
		int leftval = IsLeftValue(i, tokens1, num1);
	        char *ptr = GetSignExtensionString(varlink, 0, leftval);
	        if (ptr) AddToken(ptr, ptokenlistout);
	        if (varlink->use == CVARS_PARM || varlink->use == CVARS_LOCAL)
		{
		    AddToken(GetSpinTypeString(varlink, 0), ptokenlistout);
		    AddToken("[", ptokenlistout);
		    AddToken("@", ptokenlistout);
                    AddToken(varlink->str1, ptokenlistout);
		    AddToken("]", ptokenlistout);
		}
		else
		{
                    AddToken(varlink->str1, ptokenlistout);
		}
	    }
	    else
                AddToken(varlink->str1, ptokenlistout);

	    prev_var = 1;
	    indirect = 0;
	    continue;
	}
	else if ((coninfo = (ConInfoT *)ListFindLinkString((LinkT *)conlist, tokens1[i])))
	{
            AddToken(coninfo->str1, ptokenlistout);
	    prev_var = 1;
	    indirect = 0;
	    continue;
	}
	else if (tokens1[i][0] == '"')
	{
            AddToken("string", ptokenlistout);
            AddToken("(", ptokenlistout);
            AddToken(tokens1[i], ptokenlistout);
            AddToken(")", ptokenlistout);
	    prev_var = 1;
	    indirect = 0;
	    continue;
	}
	else if (strcmp(tokens1[i], "return") == 0)
	{
            AddToken("return", ptokenlistout);
	    prev_var = 0;
	    indirect = 0;
	    continue;
	}
	//printf("Convert %s\n", tokens1[i]);
	if (SearchList(tokens1[i], operator, SEARCH_START) >= 0) prev_var = 0;
	else if (tokens1[i][0] == ',') prev_var = 0;
	else if (tokens1[i][0] == '=') prev_var = 0;
	else if (tokens1[i][0] == '(') prev_var = 0;
	else if (tokens1[i][0] == ')') prev_var = 1;
	else if (tokens1[i][0] == ']') prev_var = 1;
	else prev_var = 1;
	indirect = 0;
	if (tokens1[i][0] == ';')
	{
	    if (i < num1 - 1) AddToken("\n", ptokenlistout);
	    continue;
	}
        AddToken(TranslateCStrToSpinStr(tokens1[i], clinetype), ptokenlistout);
    }
    return i;
}

void ConvertToSpin(LinkT **ptokenlistin, LinkT **ptokenlistout, int clinetype)
{
    int i;
    char *tokens1[200];
    int num1 = ConvertTokenListToArray(*ptokenlistin, tokens1, 200);

    if (num1 <= 0) return;

    if (tokens1[0] == 0)
    {
	AddToken("BAAAADDDDDDDDDDDDDDDD", ptokenlistout);
	AddToken("\n", ptokenlistout);
	return;
	fprintf(outfile, "BADDDDDDDD\n");
	CspinExit(1);
    }

    if (clinetype == CLINE_DIRECTIVE)
    {
#if 0
	if (prevtype == CLINE_FUNCTION || prevtype == CLINE_DECL_LOCAL)
	    AddToken("\n", ptokenlistout);
#endif
	ConvertDirectiveToSpin(tokens1, num1, ptokenlistout, clinetype);
    }
    else if (clinetype == CLINE_BRACE_LEFT)
    {
        return;
    }
    else if (clinetype == CLINE_STRUCT_TYPE)
    {
        return;
    }
    else if (clinetype == CLINE_BRACE_RIGHT)
    {
	if (modeflag == MODEFLAG_STRUCT)
	    modeflag = MODEFLAG_STRUCT1;
	else if (modeflag == MODEFLAG_TYPE_STRUCT)
	    modeflag = MODEFLAG_TYPE_STRUCT1;
	CheckEndPtr(ptokenlistout, 1);
        return;
    }
    else if (clinetype == CLINE_CASE)
    {
	if (strcmp(tokens1[0], "default") == 0)
	{
	    AddToken("OTHER", ptokenlistout);
	}
        for (i = 1; i < num1; i++)
        {
            AddToken(TranslateCStrToSpinStr(tokens1[i], clinetype), ptokenlistout);
        }
        return;
    }
    else if (clinetype == CLINE_TYPEDEF)
    {
        return;
    }
    else if (clinetype == CLINE_PROTOTYPE)
    {
        return;
    }
    else if (clinetype == CLINE_FUNCTION)
    {
	ConvertFunctionToSpin(tokens1, num1, ptokenlistout, clinetype);
    }
    else if (clinetype == CLINE_DECL_STRUCT || clinetype == CLINE_STRUCT_VAR)
    {
        ConvertDeclStructToSpin(tokens1, num1, ptokenlistout, clinetype);
    }
    else if (clinetype == CLINE_STRUCT || clinetype == CLINE_TYPE_STRUCT)
    {
        ConvertStructToSpin(tokens1, num1, ptokenlistout, clinetype);
    }
    else if (clinetype == CLINE_STRUCT_ELEM)
    {
        ConvertStructElemToSpin(tokens1, num1, ptokenlistout, clinetype);
    }
    else if (clinetype == CLINE_DECL_LOCAL)
    {
        ConvertDeclLocalToSpin(tokens1, num1, ptokenlistout, clinetype);
    }
    else if (clinetype == CLINE_DECL_GLOBAL)
    {
        ConvertDeclGlobalToSpin(tokens1, num1, ptokenlistout, clinetype);
    }
    else if (clinetype == CLINE_STATEMENT)
    {
#if 0
	if (prevtype == CLINE_FUNCTION || prevtype == CLINE_DECL_LOCAL)
	    AddToken("\n", ptokenlistout);
#endif
	ConvertStatementToSpin(0, tokens1, num1, ptokenlistout, clinetype);
    }
    else if (clinetype == CLINE_CONTROL || clinetype == CLINE_LOOP)
    {
#if 0
	if (prevtype == CLINE_FUNCTION || prevtype == CLINE_DECL_LOCAL)
	    AddToken("\n", ptokenlistout);
#endif
	ConvertControlToSpin(tokens1, num1, ptokenlistout, clinetype);
    }
    else if (clinetype == CLINE_COMMENT || clinetype == CLINE_C_COMMENT || clinetype == CLINE_C_COMMENTS)
    {
#if 0
	if (prevtype == CLINE_FUNCTION || prevtype == CLINE_DECL_LOCAL)
	    AddToken("\n", ptokenlistout);
#endif
        AddToken(TranslateCStrToSpinStr(tokens1[0], clinetype), ptokenlistout);
        //AddToken("\n", ptokenlistout);
    }
    else
    {
#if 0
	if (prevtype == CLINE_FUNCTION || prevtype == CLINE_DECL_LOCAL)
	    AddToken("\n", ptokenlistout);
#endif
        for (i = 0; i < num1; i++)
        {
	    if (tokens1[i][0] == ';')
	    {
	        if (i == num1 - 1) break;
	        if (i < num1 - 1 && tokens1[i+1][0] == '\n') continue;
	    }
	    AddToken(TranslateCStrToSpinStr(tokens1[0], clinetype), ptokenlistout);
        }
    }
    if (clinetype != CLINE_BRACE_LEFT && clinetype != CLINE_BRACE_RIGHT)
        prevtype = clinetype;
}

char *TranslateCStrToSpinStr(char *str, int clinetype)
{
    int i;
    static char buffer[CLINE_MAX_SIZE];

    strcpy(buffer, str);
    str = buffer;

    i = SearchList(str, cstr, SEARCH_EXACT);
    if (i >= 0) str = sstr[i];
    else if (str[0] == '0' && (str[1] == 'x' || str[1] == 'X'))
    {
	str++;
	*str = '$';
    }
    else if (isdigit(str[0]))
    {
	int len = strlen(str);
	if (str[len-1] == 'L') str[len-1] = 0;
    }
    else
    {
        if (str[0] == '\'')
        {
            str[0] = '"';
            str[strlen(str)-1] = '"';
        }
        else if (str[0] == '/' && str[1] == '/' )
        {
            if (!strncmp(str, "// SPIN OBJECT", 14))
            {
                str = &buffer[14 - 5];
                memcpy(str, "OBJ\n ", 5);
                return str;
            }
            else
            {
                str[1] = '\'';
                str++;
            }
        }
        else if (str[0] == '/' && str[1] == '*' )
        {
            if (!strcmp(str, "/* INLINE SPIN"))
            {
                inlineflag = 1;
                str[1] = '\'';
            }
            else
                str[1] = '{';
            str++;
        }
	else if (str[0] == '0' && (str[1] == 'x' || str[1] == 'X'))
	{
	    str++;
	    *str = '$';
	}
	if (clinetype == CLINE_C_COMMENT)
	{
	    int len = strlen(str);
	    if (str[len-2] == '*' && str[len-1] == '/')
	    {
                if (!inlineflag)
                {
		    str[len-2] = '}';
		    str[len-1] = 0;
                }
                else
                {
		    str[len-2] = 0;
                }
	    }
	}
    }
    return str;
}
