#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include "links.h"
#include "strsubs.h"
#include "vars.h"
#include "spinit.h"
#include "spinstruct.h"
#include "spinvars.h"
#include "cline.h"
#include "tokens.h"

extern VarInfoT *varlist;
extern VarInfoT *StructElem;

extern int modeflag;
extern TypeInfoT *typeslist;
extern TypeInfoT *structlink;
extern int nonewline;

void GenerateLocalSpinStructDeclaration(char *name, int size, LinkT **ptokenlistout)
{
    VarInfoT *varinfo = FindVariable(varlist, name);
    int ptrlevel = 0;
    int arraysize = 0;

    if (varinfo)
    {
	name = varinfo->str1;
	ptrlevel = varinfo->ptrlevel;
	if (varinfo->array)
	{
	    arraysize = varinfo->array;
	    ptrlevel--;
	    if (ptrlevel > 0)
	        size = 4 * arraysize;
	    else
	        size *= arraysize;
	}
	else if (ptrlevel > 0)
	    size = 1;
    }

    AddToken(name, ptokenlistout);
    if (size != 1)
    {
        char *ptr = str_alloc(20);
        sprintf(ptr, "%d", size >> 2);
        AddToken("[", ptokenlistout);
        AddToken(ptr, ptokenlistout);
        AddToken("]", ptokenlistout);
    }
}

void GenerateGlobalSpinStructDeclaration(char *name, int size, LinkT **ptokenlistout)
{
    AddToken("  long", ptokenlistout);
    GenerateLocalSpinStructDeclaration(name, size, ptokenlistout);
}

void ConvertDeclStructToSpin(char **tokens1, int num1, LinkT **ptokenlistout, int clinetype)
{
    int i;
    int size = 0;
    TypeInfoT *typeinfo;

    if (clinetype == CLINE_STRUCT_VAR)
    {
	i = 0;
        typeinfo = structlink;
    }
    else
    {
	i = 2;
        typeinfo= FindType(typeslist, tokens1[1]);
    }

    if (tokens1[i][0] == '*') i++;

    if (typeinfo) size = typeinfo->size;

    if (modeflag != MODEFLAG_VAR)
    {
        AddToken("\n", ptokenlistout);
        AddToken("VAR", ptokenlistout);
        AddToken("\n", ptokenlistout);
        modeflag = MODEFLAG_VAR;
    }
    GenerateGlobalSpinStructDeclaration(tokens1[i], size, ptokenlistout);
    //if (!nonewline) AddToken("\n", ptokenlistout);
}

void ConvertStructToSpin(char **tokens1, int num1, LinkT **ptokenlistout, int clinetype)
{
    int i;
    if (modeflag != MODEFLAG_CON)
    {
        AddToken("\n", ptokenlistout);
        AddToken("CON", ptokenlistout);
        AddToken("\n", ptokenlistout);
        modeflag = MODEFLAG_CON;
    }
    AddToken("'", ptokenlistout);
    for (i = 0; i < num1; i++) AddToken(tokens1[i], ptokenlistout);
}

void ConvertStructElemToSpin(char **tokens1, int num1, LinkT **ptokenlistout, int clinetype)
{
    int i;
    if (modeflag != MODEFLAG_CON)
    {
        AddToken("\n", ptokenlistout);
        AddToken("CON", ptokenlistout);
        AddToken("\n", ptokenlistout);
        modeflag = MODEFLAG_CON;
    }
    for (i = 0; i < num1; i++) if (tokens1[i][0] == ';') break;
    if (i > 0 && i < num1)
    {
        char *ptr;
	if (StructElem) AddToken(StructElem->str1, ptokenlistout);
	AddToken("=", ptokenlistout);
	ptr = str_alloc(10);
	if (StructElem) sprintf(ptr, "%d", StructElem->value);
	else strcpy(ptr, "????");
	AddToken(ptr, ptokenlistout);
        //if (!nonewline) AddToken("\n", ptokenlistout);
    }
}

int ConvertStructReferenceToSpin(int i, char **tokens1, int num1, LinkT **ptokenlistout, int indirect)
{
    //int j;
    //int num20 = num2;
    VarInfoT *varlink;
    TypeInfoT *typeinfo;
    VarInfoT *structvars;
    VarInfoT *varinfo;
    int ptrlevel;
    LinkT *token20;
    int refstructptr = 0;

    varlink = FindVariable(varlist, tokens1[i]);
    if (varlink == 0) return i;
    varinfo = varlink;

    // Check if just struct name alone
    if (tokens1[i+1][0] != '[' &&
        tokens1[i+1][0] != '.' &&
	strcmp(tokens1[i+1], "->") != 0)
    {
	AddToken(varinfo->str1, ptokenlistout);
	return i;
    }

    //token20 = AddToken("????", ptokenlistout);
    token20 = AddToken("[", ptokenlistout);

    // Determine if pointer or not
#if 1
    ptrlevel = varlink->ptrlevel;
    if (varlink->array) ptrlevel--;
    if (ptrlevel < 1)
        AddToken("@", ptokenlistout);
#else
    j = i + 1;
    if (tokens1[j][0] == '[')
    {
	while (tokens1[j++][0] != ']')
	{
	    if (j >= num1) break;
	    j++;
	}
    }

    if (tokens1[j][0] == '.')
        AddToken("@", ptokenlistout);
#endif
    AddToken(varlink->str1, ptokenlistout);

    while (i < num1 - 2 && (tokens1[i+1][0] == '.' ||
        strcmp(tokens1[i+1], "->") == 0 || tokens1[i+1][0] == '['))
    {
	if (refstructptr)
	{
	    refstructptr = 0;
	    AddToken("]", ptokenlistout);
	    token20 = InsertToken("long", ptokenlistout, token20);
	    token20 = InsertToken("[", ptokenlistout, token20);
	}
        typeinfo = varlink->typeinfo;
        structvars = (VarInfoT *)typeinfo->structvars;
        if (structvars == 0)
        {
	    if (tokens1[i+1][0] != '[')
	    {
	        printf("**** NO STRUCT VARS ****\n");
	        AddToken("ERROR", ptokenlistout);
	    }
	    break;
        }
	if (tokens1[i+1][0] == '[')
	{
	    char *str = str_alloc(15);
	    int parenflag = tokens1[i+3][0] != ']';
	    int size = typeinfo->size;
	    if (varlink->ptrlevel > 1) size = 4;
	    sprintf(str, "%d", size);
            AddToken("+", ptokenlistout);
            AddToken("(", ptokenlistout);
            AddToken(str, ptokenlistout);
            AddToken("*", ptokenlistout);
	    if (parenflag) AddToken("(", ptokenlistout);
	    i += 2;
	    while (i < num1 && tokens1[i][0] != ']')
                AddToken(tokens1[i++], ptokenlistout);
	    if (parenflag) AddToken(")", ptokenlistout);
            AddToken(")", ptokenlistout);
	    continue;
	}
        varinfo = FindVariable(structvars, tokens1[i+2]);
        if (varinfo == 0)
        {
	    PrintFileLine();
	    printf(" - **** %s NOT AN ELEMENT OF STRUCT %s ****\n",
	        tokens1[i+2], varlink->str);
	    CspinExit(1);
        }
        AddToken("+", ptokenlistout);
        AddToken(varinfo->str1, ptokenlistout);
	varlink = varinfo;
        i += 2;
	// Check if this is a struct pointer
	if (varinfo->ptrlevel < 1) continue;
	if (varinfo->ptrlevel < 2 && varinfo->array) continue;
	if (varinfo->typeinfo == 0) continue;
	if (varinfo->typeinfo->structvars == 0) continue;
	refstructptr = 1;
    }

    if (tokens1[i+1][0] == '[') indirect--;
    ptrlevel = varinfo->ptrlevel + indirect;
    if (ptrlevel > 0 && varinfo->array)
    {
	ReplaceToken("(", ptokenlistout, token20);
        AddToken(")", ptokenlistout);
    }
    else
    {
	int leftval = IsLeftValue(i, tokens1, num1);
	char *ptr = GetSignExtensionString(varinfo, indirect, leftval);
        AddToken("]", ptokenlistout);
	token20 = InsertToken(GetSpinTypeString(varinfo, indirect), ptokenlistout, token20);
	if (ptr) token20 = InsertToken(ptr, ptokenlistout, token20);
    }
    return i;
}

int ConvertPointerReferenceToSpin(int i, char **tokens1, int num1, LinkT **ptokenlistout, int indirect)
{
    int j;
    char *ptr;
    int leftval;
    int ptrlevel = 0;
    int arrayindex[2];
    int arrayindex1[2];
    int arraycount = 0;
    int bracketflag = 0;
    VarInfoT *varinfo = FindVariable(varlist, tokens1[i]);

    // Check the variable info pointer
    if (varinfo == 0)
    {
	AddToken(tokens1[i], ptokenlistout);
	return i;
    }

    // Locate and count the array indices
    for (i++; i < num1; i++)
    {
	if (bracketflag == 0)
	{
	    if (tokens1[i][0] != '[')
	    {
		i--;
		break;
	    }
	    bracketflag = 1;
	    arrayindex[arraycount] = i;
	}
	else if (tokens1[i][0] == '[') bracketflag++;
	else if (tokens1[i][0] == ']')
	{
	    if (--bracketflag > 0) continue;
	    arrayindex1[arraycount++] = i;
	    if (arraycount >= 2) break;
	}
    }

    // Add an indirect index if -1
    if (indirect == -1 && arraycount < 2)
    {
	arrayindex[arraycount++] = -1;
	//printf("Add extra arrayindex\n");
    }

    // Determine the pointer level
    ptrlevel = varinfo->ptrlevel;
    if (varinfo->array) ptrlevel--;

    if (arraycount == 0)
    {
	if (varinfo->array) AddToken("@", ptokenlistout);
	AddToken(varinfo->str1, ptokenlistout);
	return i;
    }
    else if (arraycount == 1)
    {
	int rawflag = varinfo->array && (varinfo->ptrlevel >= 2 ||
	    varinfo->use == CVARS_GLOBAL ||
            (varinfo->typeinfo && varinfo->typeinfo->size == 4));

        if (!rawflag)
	{
	    leftval = IsLeftValue(i, tokens1, num1);
	    ptr = GetSignExtensionString(varinfo, -1, leftval);
	    if (ptr) AddToken(ptr, ptokenlistout);
            AddToken(GetSpinTypeString(varinfo, -1), ptokenlistout);
            AddToken("[", ptokenlistout);
	    if (varinfo->array)
	        AddToken("@", ptokenlistout);
	}
        AddToken(varinfo->str1, ptokenlistout);

	if (arrayindex[0] < 0 && i < num1 - 1 &&
	    (strcmp(tokens1[i+1], "++") == 0 ||
            strcmp(tokens1[i+1], "--") == 0))
            AddToken(tokens1[++i], ptokenlistout);

        if (!rawflag)
            AddToken("]", ptokenlistout);

	if (arrayindex[0] >= 0)
	{
            AddToken("[", ptokenlistout);
	    //printf("Call ConvertStatement\n");
            ConvertStatementToSpin(arrayindex[0]+1, tokens1, arrayindex1[0], ptokenlistout, CLINE_STATEMENT);
	    i = arrayindex1[0];
            AddToken("]", ptokenlistout);
	}
    }
    else
    {
        leftval = IsLeftValue(i, tokens1, num1);
	ptr = GetSignExtensionString(varinfo, -2, leftval);
	if (ptr) AddToken(ptr, ptokenlistout);
        AddToken(GetSpinTypeString(varinfo, -2), ptokenlistout);
        AddToken("[", ptokenlistout);

	if (!varinfo->array)
	{
            AddToken("long", ptokenlistout);
            AddToken("[", ptokenlistout);
            AddToken(varinfo->str1, ptokenlistout);
            AddToken("]", ptokenlistout);
	}
	else
            AddToken(varinfo->str1, ptokenlistout);

        for (j = arrayindex[0]; j < num1; j++)
        {
            AddToken(tokens1[j], ptokenlistout);
            if (tokens1[j][0] == ']') break;
        }

        AddToken("]", ptokenlistout);
	if (arrayindex[1] >= 0)
	{
            AddToken("[", ptokenlistout);
            ConvertStatementToSpin(arrayindex[1]+1, tokens1, arrayindex1[1], ptokenlistout, CLINE_STATEMENT);
	    i = arrayindex1[1];
            AddToken("]", ptokenlistout);
	}
    }

    return i;
}
