#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include "links.h"
#include "strsubs.h"
#include "vars.h"
#include "spinit.h"
#include "spinstruct.h"
#include "spinvars.h"
#include "cline.h"
#include "tokens.h"
#include "cdirect.h"

extern int nonewline;
static char *spintype[] = {"byte", "word", "long", "long"};
static char *typelist[] = {"char", "short", "int", "long", "void", "FILE", "static", "extern", "unsigned", "signed", "float", "double", 0};
extern int modeflag;
extern int prevtype;
extern TypeInfoT *typeslist;
extern VarInfoT *varlist;
extern ConInfoT *conlist;

char *GetSignExtensionString(VarInfoT *varlink, int indirect, int leftval)
{
    char *ptr = 0;
    int size = GetVarSizeLink(varlink, indirect);

    if (leftval == 0 && varlink->signflag == 1)
    {
        if (size == 1)
	    ptr = "~";
        else if (size == 2)
            ptr = "~~";
    }

    return ptr;
}

char *GetSpinTypeString(VarInfoT *varlink, int indirect)
{
    char *ptr;
    int size = GetVarSizeLink(varlink, indirect);

    if (size == 1)
        ptr = "byte";
    else if (size == 2)
        ptr = "word";
    else if (size == 4)
        ptr = "long";
    else if (size == 8)
        ptr = "long";
    else
        ptr = "long";

    return ptr;
}

void ConvertDirectiveToSpin(char **tokens1, int num1, LinkT **ptokenlistout, int clinetype)
{
    int i;
    char *name;
    ConInfoT *coninfo;

    if (strcmp(tokens1[0], "#") == 0 && strcmp(tokens1[1], "macro") == 0) return;

#if 0
    if (prevtype == CLINE_FUNCTION || prevtype == CLINE_DECL_LOCAL)
        AddToken("\n", ptokenlistout);
#endif

    if (num1 > 0 && tokens1[num1-1][0] == '\n') num1--;

    if (num1 >= 3 && strcmp(tokens1[0], "#") == 0 && strcmp(tokens1[1], "define") == 0)
    {
	if (num1 < 4) return;

	name = tokens1[2];
        coninfo = (ConInfoT *)ListFindLinkString((LinkT *)conlist, name);

	if (coninfo == 0)
	{
	    PrintFileLine();
	    printf("Could not find %s in conlist\n", name);
	    return;
	}

        if (coninfo->type == CONTYPE_STRING) return;

        name = coninfo->str1;

        if (modeflag != MODEFLAG_CON)
        {
	    AddToken("\n", ptokenlistout);
	    AddToken("CON", ptokenlistout);
	    AddToken("\n", ptokenlistout);
	    modeflag = MODEFLAG_CON;
        }

        AddToken(" ", ptokenlistout);
        AddToken(name, ptokenlistout);
        AddToken("=", ptokenlistout);
	for (i = 3; i < num1; i++)
	{
	    if (tokens1[i][0] == '\n') break;
            AddToken(TranslateCStrToSpinStr(tokens1[i], 0), ptokenlistout);
	}
        //if (!nonewline) AddToken("\n", ptokenlistout);
    }
    else
    {
        AddToken("'CSPIN WARNING", ptokenlistout);
	for (i = 0; i < num1; i++)
            AddToken(tokens1[i], ptokenlistout);
#if 0
	if (tokens1[num1-1][0] != '\n')
            AddToken("\n", ptokenlistout);
#endif
    }
}

void ConvertDeclLocalToSpin(char **tokens1, int num1, LinkT **ptokenlistout, int clinetype)
{
    int i;
    int firstflag = 1;
    TypeInfoT *typeinfo;
    VarInfoT *varinfo;
    int bracketflag;
    int size;

    for (i = 0; i < num1; i++)
    {
        if (tokens1[i][0] == '\n') break;
	if ((typeinfo = FindType(typeslist, tokens1[i])))
	{
	    if (typeinfo->structvars == 0 || tokens1[i+1][0] == '*') continue;
            if (firstflag)
            {
	        firstflag = 0;
	        if (prevtype == CLINE_FUNCTION)
	            AddToken("|", ptokenlistout);
	        else
	            AddToken(",", ptokenlistout);
            }
	    if (tokens1[i+1][0] == '*') i++;
            GenerateLocalSpinStructDeclaration(tokens1[i+1], typeinfo->size, ptokenlistout);
	    i++;
	    continue;
	}
        if (SearchList(tokens1[i], typelist, SEARCH_EXACT) >= 0) continue;
        if (tokens1[i][0] == '*') continue;
        if (tokens1[i][0] == ';') continue;
	if (tokens1[i][0] == ',')
	{
            AddToken(tokens1[i], ptokenlistout);
	    continue;
	}

        if (firstflag)
        {
	    firstflag = 0;
	    if (prevtype == CLINE_FUNCTION)
	        AddToken("|", ptokenlistout);
	    else
	        AddToken(",", ptokenlistout);
        }

	if ((varinfo = FindVariable(varlist, tokens1[i])) == 0) continue;

        AddToken(varinfo->str1, ptokenlistout);
        if (varinfo->array == 0) continue;

        // Copy array info
	i++;

	if (varinfo->ptrlevel < 2 && varinfo->typeinfo)
	    size = varinfo->typeinfo->size;
	else
	    size = 4;

	for (bracketflag = 0; i < num1; i++)
	{
	    if (!bracketflag)
	    {
		if (tokens1[i][0] != '[') break;
		if (isdigits(tokens1[i+1]) && tokens1[i+2][0] == ']')
		{
		    int value;
		    char tempstr[20];
		    sscanf(tokens1[i+1], "%d", &value);
		    if (size == 1) value = (value + 3) / 4;
		    else if (size == 2) value = (value + 1) / 2;
		    sprintf(tempstr, "%d", value);
		    AddToken("[", ptokenlistout);
		    AddToken(tempstr, ptokenlistout);
		    AddToken("]", ptokenlistout);
		    i += 2;
		}
		else
		{
		    bracketflag = 1;
		    AddToken(tokens1[i], ptokenlistout);
		    if (size == 1 || size == 2)
		        AddToken("(", ptokenlistout);
		}
		continue;
	    }
	    else if (tokens1[i][0] == ']')
	    {
		if (size == 1)
		{
		    AddToken("+", ptokenlistout);
		    AddToken("3", ptokenlistout);
		    AddToken(")", ptokenlistout);
		    AddToken("/", ptokenlistout);
		    AddToken("4", ptokenlistout);
		    AddToken(")", ptokenlistout);
		}
		else if (size == 2)
		{
		    AddToken("+", ptokenlistout);
		    AddToken("1", ptokenlistout);
		    AddToken(")", ptokenlistout);
		    AddToken("/", ptokenlistout);
		    AddToken("2", ptokenlistout);
		    AddToken(")", ptokenlistout);
		}
		bracketflag = 0;
	    }
	    AddToken(tokens1[i], ptokenlistout);
	}
	i--;
    }
}

void ConvertDeclGlobalToSpin(char **tokens1, int num1, LinkT **ptokenlistout, int clinetype)
{
    int i;

    // Look for an equals sign
    for (i = 0; i < num1; i++)
    {
	if (strcmp(tokens1[i], "=") == 0) break;
    }

    if (i < num1)
        ConvertDeclGlobalDatToSpin(tokens1, num1, ptokenlistout, clinetype);
    else
        ConvertDeclGlobalVarToSpin(tokens1, num1, ptokenlistout, clinetype);

}

void ConvertDeclGlobalVarToSpin(char **tokens1, int num1, LinkT **ptokenlistout, int clinetype)
{
    int i, k;
    int pointer = 0;
    TypeInfoT *typeinfo;
    TypeInfoT *typetemp;

    if (modeflag != MODEFLAG_VAR)
    {
        AddToken("\n", ptokenlistout);
        AddToken("VAR", ptokenlistout);
        AddToken("\n", ptokenlistout);
        modeflag = MODEFLAG_VAR;
    }

    typeinfo = FindType(typeslist, "int");
    k = 0;
    for (i = 0; i < num1; i++)
    {
        if (tokens1[i][0] == '*')
        {
	    pointer = 1;
	    continue;
	}
	else if ((typetemp = FindType(typeslist, tokens1[i])))
	{
	    typeinfo = typetemp;
	    if (typeinfo->structvars)
	    {
		if (tokens1[i+1][0] == '*') i++;
                GenerateGlobalSpinStructDeclaration(tokens1[i+1], typeinfo->size, ptokenlistout);
                //AddToken("\n", ptokenlistout);
		return;
	    }
	}
	else if (FindVariable(varlist, tokens1[i]))
	    break;
    }
    AddToken(" ", ptokenlistout);
    if (pointer || typeinfo->ptrlevel)
        AddToken("long", ptokenlistout);
    else if (typeinfo->size == 1)
        AddToken("byte", ptokenlistout);
    else if (typeinfo->size == 2)
        AddToken("word", ptokenlistout);
    else if (typeinfo->size == 4)
        AddToken("long", ptokenlistout);
    else
        AddToken("????", ptokenlistout);

    for (; i < num1; i++)
    {
	if (tokens1[i][0] == ';') continue;
	if (tokens1[i][0] == ',')
	    AddToken(tokens1[i], ptokenlistout);
	else
	    AddToken(GetSpinVariableName(tokens1[i]), ptokenlistout);
    }

}

static char *skiplist[] = {"extern", "static", "unsigned", 0};

void ConvertDeclGlobalDatToSpin(char **tokens1, int num1, LinkT **ptokenlistout, int clinetype)
{
    int i, j, k;
    int strcount = 0;
    int savei;

    if (modeflag != MODEFLAG_DAT)
    {
        AddToken("\n", ptokenlistout);
        AddToken("DAT", ptokenlistout);
        AddToken("\n", ptokenlistout);
        modeflag = MODEFLAG_DAT;
    }
    // Skip "static", "extern" or "unsigned"
    for (i = 0; i < num1; i++)
    {
	if (SearchList(tokens1[i], skiplist, SEARCH_EXACT) < 0) break;
    }

    savei = i;

    if (strcmp(tokens1[i], "char") == 0 && strcmp(tokens1[i+1], "*") == 0 && strcmp(tokens1[i+3], "=") == 0 && tokens1[i+4][0] == '"')
    {
	AddToken(" ", ptokenlistout);
	AddToken(GetSpinVariableName(tokens1[i+2]), ptokenlistout);
	AddToken("byte", ptokenlistout);
	AddToken(tokens1[i+4], ptokenlistout);
	AddToken(",", ptokenlistout);
	AddToken("0", ptokenlistout);
        //AddToken("\n", ptokenlistout);
    }

    // Check for array of pointers to strings
    else if (strcmp(tokens1[i], "char") == 0 && strcmp(tokens1[i+1], "*") == 0 && strcmp(tokens1[i+3], "[") == 0 &&
        ((strcmp(tokens1[i+4], "]") == 0 && strcmp(tokens1[i+5], "=") == 0 && strcmp(tokens1[i+6], "{") == 0) ||
        (strcmp(tokens1[i+5], "]") == 0 && strcmp(tokens1[i+6], "=") == 0 && strcmp(tokens1[i+7], "{") == 0)))
    {
        int offset[100];
        int n = 0;
        int m = 0;
        int mm = 0;
        int first = 1;
	char *numptr;
	char *strptr2;
	char *strptr = AddToSymbolList("datstr");

	AddToken(" ", ptokenlistout);
	AddToken(strptr, ptokenlistout);
	AddToken("byte", ptokenlistout);
	mm += strlen(strptr) + 7;
        if (tokens1[i+5][0] == '=') i += 7;
        else i += 8;
        for (; i < num1; i++)
        {
            if (strcmp(tokens1[i], ",") == 0) continue;
            if (strcmp(tokens1[i], "}") == 0) break;
            if (tokens1[i][0] == '"')
            {
                if (mm >= 50)
                {
                    mm = 0;
                    first = 1;
	            AddToken("\n", ptokenlistout);
	            AddToken(" ", ptokenlistout);
	            AddToken("byte", ptokenlistout);
                }
                if (first) first = 0;
                else AddToken(",", ptokenlistout);
                offset[n++] = m;
                AddToken(tokens1[i], ptokenlistout);
                AddToken(",", ptokenlistout);
                AddToken("0", ptokenlistout);
                m += strlen(tokens1[i]) - 1;
                mm += strlen(tokens1[i]) + 5;
            }
            else
            {
                offset[n++] = -1;
            }
        }
        mm = 0;
        first = 1;
        AddToken("\n", ptokenlistout);
        AddToken(" ", ptokenlistout);
	strptr2 = GetSpinVariableName(tokens1[savei+2]);
        AddToken(strptr2, ptokenlistout);
        AddToken("long", ptokenlistout);
	mm += strlen(strptr2) + 7;
	//printf("n = %d, offset[0] = %d\n", n, offset[0]);
        for (i = 0; i < n; i++)
        {
            if (mm >= 60)
            {
                mm = 0;
                first = 1;
                AddToken("\n", ptokenlistout);
                AddToken("  long", ptokenlistout);
		mm += 6;
            }
            if (first) first = 0;
            else AddToken(",", ptokenlistout);
            if (offset[i] >= 0)
            {
                AddToken("@", ptokenlistout);
                AddToken(strptr, ptokenlistout);
                AddToken("+", ptokenlistout);
		numptr = str_alloc(15);
		sprintf(numptr, "%d", offset[i] + 16);
                AddToken(numptr, ptokenlistout);
                mm += 15;
                if (strcount >= 10) mm++;
                if (i >= 10) mm++;
            }
            else
            {
                AddToken("0", ptokenlistout);
                mm += 3;
            }
        }
        //AddToken("\n", ptokenlistout);
    }

    else if ((j = SearchList(tokens1[i], typelist, SEARCH_EXACT)) >= 0)
    {
	int bracketflag = 0;
	k = j;
	while (i < num1)
	{
	    if (tokens1[++i][0] == '*') continue;
            j = SearchList(tokens1[i], typelist, SEARCH_EXACT);
	    if (j < 0) break;
	    k = j;
	}
        AddToken(" ", ptokenlistout);
	AddToken(GetSpinVariableName(tokens1[i]), ptokenlistout);
	if (k >= 0 && k <= 3)
            AddToken(spintype[k], ptokenlistout);
	else
            AddToken("????", ptokenlistout);
	for (i++; i < num1; i++)
	{
	    if (bracketflag)
	    {
	        if (tokens1[i][0] == ']')
		    bracketflag = 0;
		continue;
	    }
	    if (tokens1[i][0] == '[')
	    {
		bracketflag = 1;
		continue;
	    }
	    if (tokens1[i][0] == '=') continue;
	    if (tokens1[i][0] == '{') continue;
	    if (tokens1[i][0] == '}') continue;
	    if (tokens1[i][0] == ';') continue;
	    if (tokens1[i][0] == '\n') continue;
            AddToken(tokens1[i], ptokenlistout);
	}
        //AddToken("\n", ptokenlistout);
    }

    // Otherwise, just copy the tokens
    else
    {
        for (i = 0; i < num1; i++) AddToken(tokens1[i], ptokenlistout);
    }
}
