#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include "links.h"
#include "strsubs.h"
#include "vars.h"
#include "cline.h"
#include "tokens.h"
#include "cdirect.h"

static char *special = " \t*()/\"'=+-%,;<>&|^!~#{}[]";
static char *types[] = {"LOCAL", "GLOBAL", "FUNCTION", "PARM", "STRUCT", "PROTOTYPE"};
extern TypeInfoT *typeslist;

//ConInfoT *conlist = 0;
VarInfoT *varlist = 0;
VarInfoT *lvarlist = 0;
VarInfoT *StructElem = 0;
TypeInfoT *structlink = 0;
static int structnum = 0;
extern char objname[];

int SizeOfVarInfo(VarInfoT *varinfo)
{
    int size = 0;
    int arraysize;
    int ptrlevel;
    TypeInfoT *typeinfo;

    if (varinfo)
    {
        ptrlevel = varinfo->ptrlevel;
        arraysize = varinfo->array;
	//printf("SizeOfVarInfo: %s, %d, %d\n", varinfo->str, ptrlevel, arraysize);

        if (arraysize)
            ptrlevel--;

        else
            arraysize = 1;

        if (ptrlevel > 0)
            size = 4 * arraysize;

        else if ((typeinfo = varinfo->typeinfo))
	{
            size = typeinfo->size * arraysize;
	    //printf("SizeOfVarInfo: type %s has has size %d\n",
	        //typeinfo->str, typeinfo->size);
	}
	else
	{
	    printf("SizeOfVarInfo: %s has no type\n", varinfo->str);
	    CspinExit(1);
        }
    }

    return size;
}

TypeInfoT *TypeInfoAppend(TypeInfoT *list, TypeInfoT *link)
{
    return (TypeInfoT *)ListAppend((LinkT *)list, (LinkT *)link);
}

static VarInfoT *VarInfoAppend(VarInfoT *list, VarInfoT *varinfo)
{
    return (VarInfoT *)ListAppend((LinkT *)list, (LinkT *)varinfo);
}

TypeInfoT *AddToTypesList(char *typename, int typesize, int signflag)
{
    TypeInfoT *typeinfo = (TypeInfoT *)LinkCreate(typename, sizeof(TypeInfoT));
    typeinfo->size = typesize;
    typeinfo->signflag = signflag;
    typeslist = TypeInfoAppend(typeslist, typeinfo);
    return typeinfo;
}

TypeInfoT *FindType(TypeInfoT *typeinfo, char *str)
{
    return (TypeInfoT *)ListFindLinkString((LinkT *)typeslist, str);
}

VarInfoT *FindVariable(VarInfoT *varinfo, char *varname)
{
    LinkT *varlink = (LinkT *)varinfo;

    varlink = ListFindLinkStringLast(varlink, varname);

    return (VarInfoT *)varlink;
}

void ProcessTypedef(char **tokens, int num)
{
    int i;
    int signflag = -1;
    int ptrlevel = 0;
    char *newtype = 0;
    char *basetype = "int";
    TypeInfoT *typeinfo;
    TypeInfoT *baseinfo;

    if (num < 3) return;
    if (strcmp(tokens[0], "typedef") != 0) return;
    for (i = 1; i < num; i++)
    {
	if (tokens[i][0] == ';')
	    break;
	else if (strcmp(tokens[i], "signed") == 0)
	    signflag = 1;
	else if (strcmp(tokens[i], "unsigned") == 0)
	    signflag = 0;
	else if (strcmp(tokens[i], "*") == 0)
	    ptrlevel = 1;
	else if (strcmp(tokens[i], "**") == 0)
	    ptrlevel = 2;
	else if (strcmp(tokens[i], "***") == 0)
	    ptrlevel = 3;
	else if (tokens[i+1][0] == ';')
	{
	    newtype = tokens[i];
	}
	else
	    basetype = tokens[i];

    }
    if (i < 3) return;
    if (i >= num) return;
    if (FindType(typeslist, newtype)) return;
    baseinfo = FindType(typeslist, basetype);
    if (baseinfo == 0) return;
    typeinfo = (TypeInfoT *)LinkCreate(newtype, sizeof(TypeInfoT));
    typeinfo->size = baseinfo->size;
    typeinfo->ptrlevel = baseinfo->ptrlevel + ptrlevel;
    if (signflag == 1)
        typeinfo->signflag = 1;
    else if (signflag == 0)
        typeinfo->signflag = 0;
    else
        typeinfo->signflag = baseinfo->signflag;
    typeinfo->floatflag = baseinfo->floatflag;
    typeinfo->typedefptr = baseinfo;
    typeslist = TypeInfoAppend(typeslist, typeinfo);
}

int GetArrayLink(VarInfoT *varinfo)
{
    if (varinfo == 0) return -1;
    return varinfo->array;
}

int GetVarSizeElem(VarInfoT *varinfo)
{
    TypeInfoT *typeinfo;

    if (varinfo == 0) return 0;
    typeinfo = varinfo->typeinfo;
    if (typeinfo == 0) return 0;
    return typeinfo->size;
}

int GetVarSizeLink(VarInfoT *varinfo, int indirect)
{
    if (varinfo == 0) return 0;
    if (varinfo->ptrlevel + indirect > 0) return 4;
    return GetVarSizeElem(varinfo);
}

int GetUseLink(VarInfoT *varinfo)
{
    if (varinfo == 0) return -1;
    return varinfo->use;
}

void seperateinits(LinkT **ptokenlistin, LinkT **ptokenlistdefer)
{
    char *token;
    LinkT *next;
    LinkT *list;
    LinkT *prev = *ptokenlistin;

    if (prev)
        list = prev->next;
    else
        return;

    while (list)
    {
        // Check for initialization
	if (list->str[0] == '=')
        {
            int parencount = 0;
            int bracecount = 0;

            AddToken(prev->str, ptokenlistdefer);

	    while (list)
            {
		next = list->next;
		token = list->str;
                if (parencount <= 0 && bracecount <= 0 && (token[0] == ',' || token[0] == ';'))
                {
                    AddToken(";", ptokenlistdefer);
                    //AddToken("\n", ptokenlistdefer);
                    break;
                }
                else if (token[0] == '(') parencount++;
                else if (token[0] == ')') parencount--;
                else if (token[0] == '{') bracecount++;
                else if (token[0] == '}') bracecount--;
                AddToken(token, ptokenlistdefer);
		RemoveToken(ptokenlistin, list);
		list = next;
            }
        }
	prev = list;
	if (list) list = list->next;
    }
}

int ProcessVariable(int i, int num, char **tokens)
{
    char *ptr;
    int ptrlevel = 0;
    int signflag = -1;
    TypeInfoT *vartypetemp;
    TypeInfoT *vartypeinfo = 0;
    char typestr[100];

    for (; i < num; i++)
    {
        // Skip "extern", "static" and special characters
        ptr = FindChar(special, tokens[i][0]);
        if (*ptr != 0) continue;
        if (strcmp(tokens[i], "extern") == 0) continue;
        if (strcmp(tokens[i], "static") == 0) continue;
        if (strcmp(tokens[i], "unsigned") == 0)
        {
            signflag = 0;
            continue;
        }
        if (strcmp(tokens[i], "signed") == 0)
        {
            signflag = 1;
            continue;
        }
        if (strcmp(tokens[i], "struct") == 0) continue;
        if (tokens[i][0] < ' ') continue;
    
        // Check for type keyword
        vartypetemp = FindType(typeslist, tokens[i]);
        if (vartypetemp)
        {
            vartypeinfo = vartypetemp;
            strcpy(typestr, tokens[i]);
            ptrlevel = 0;
            if (i < num - 1 && tokens[i+1][0] == '*')
            {
                i++;
                ptrlevel += strlen(tokens[i]);
                //printf("string = -%s-, strlen = %d\n", tokens[i], strlen(tokens[i]));
            }
            //printf("type = %s, ptrlevel = %d\n", typestr, ptrlevel);
            continue;
        }
        else
          break;
    }
    return i;
}

int getcvars(LinkT *tokenlist, int clinetype)
{
    char typestr[100];
    int use;
    int i;
    char *ptr;
    int vcount = 0;
    int ptrlevel = 0;
    int signflag = -1;
    TypeInfoT *vartypetemp;
    TypeInfoT *vartypeinfo = 0;
    VarInfoT *varinfo;
    int bracketflag;
    int firsti, temp;
    char *tokens[200];
    int num = ConvertTokenListToArray(tokenlist, tokens, 200);

    if (clinetype == CLINE_TYPEDEF)
    {
	ProcessTypedef(tokens, num);
	return clinetype;
    }
    else if (clinetype == CLINE_TYPE_STRUCT)
    {
	char *name;
	if (num < 3 || tokens[2][0] == '\n')
	    name = AddToSymbolList("_struct");
	else name = tokens[2];
	structlink = AddToTypesList(name, 0, 0);
	if (structlink) structlink->value = ++structnum;
    }
    else if (clinetype == CLINE_STRUCT)
    {
	char *name;
	if (num < 2 || tokens[1][0] == '\n')
	    name = AddToSymbolList("_struct");
	else name = tokens[1];
	structlink = AddToTypesList(name, 0, 0);
	if (structlink) structlink->value = ++structnum;
    }
    else if (clinetype == CLINE_STRUCT_TYPE)
    {
	for (i = 0; i < num; i++) if (tokens[i][0] == ';') break;
	//printf("CLINE_STRUCT_TYPE: %s\n", tokens[i-1]);
	if (i > 0 && i < num)
	{
	    TypeInfoT *typeinfo = AddToTypesList(tokens[i-1], 0, 0);
	    if (structlink)
	    {
		// Round size to nearest multiple of 4
		structlink->size = (structlink->size + 3) & ~3;

		// Inherent size and variable list
		typeinfo->typedefptr = structlink;
		typeinfo->size = structlink->size;
		typeinfo->structvars = structlink->structvars;
	    }
	}
	return clinetype;
    }

    if (clinetype == CLINE_STRUCT_VAR)
    {
	//printf("CLINE_STRUCT_VAR:\n");
	// Round size to nearest multiple of 4
        if (structlink)
	{
	    structlink->size = (structlink->size + 3) & ~3;
	    strcpy(typestr, structlink->str);
	    vartypeinfo = structlink;
	}
    }
    else if (clinetype == CLINE_FUNCTION || clinetype == CLINE_DECL_LOCAL ||
        clinetype == CLINE_DECL_GLOBAL || clinetype == CLINE_DECL_STRUCT ||
        clinetype == CLINE_STRUCT_ELEM )
    {
        strcpy(typestr, "int");
    }
    else if (clinetype == CLINE_PROTOTYPE)
    {
        strcpy(typestr, "int");
        // Get type information and locate function name
        //printf("Prototype\n");
        i = ProcessVariable(0, num, tokens);
        varinfo = AddObjFuncProtoToVarList(tokens[i]);
        return 0;
    }
    else
        return 0;

    for (i = 0; i < num; i++)
    {
        // Skip over initialization
        if (tokens[i][0] == '=')
        {
            int parencount = 0;
            int bracecount = 0;
            while (++i < num)
            {
                if (parencount <= 0 && bracecount <= 0 && tokens[i][0] == ',') break;
                else if (tokens[i][0] == '(') parencount++;
                else if (tokens[i][0] == ')') parencount--;
                else if (tokens[i][0] == '{') bracecount++;
                else if (tokens[i][0] == '}') bracecount--;
            }
            continue;
        }

        // Skip "extern", "static" and special characters
        ptr = FindChar(special, tokens[i][0]);
        if (*ptr != 0) continue;
        if (strcmp(tokens[i], "extern") == 0) continue;
        if (strcmp(tokens[i], "static") == 0) continue;
        if (strcmp(tokens[i], "unsigned") == 0)
	{
	    signflag = 0;
	    continue;
	}
        if (strcmp(tokens[i], "signed") == 0)
	{
	    signflag = 1;
	    continue;
	}
        if (strcmp(tokens[i], "struct") == 0) continue;
	if (tokens[i][0] < ' ') continue;

        // Check for type keyword
	vartypetemp = FindType(typeslist, tokens[i]);
	if (vartypetemp)
        {
	    vartypeinfo = vartypetemp;
            strcpy(typestr, tokens[i]);
            ptrlevel = 0;
            if (i < num - 1 && tokens[i+1][0] == '*')
            {
                i++;
		ptrlevel += strlen(tokens[i]);
		//printf("string = -%s-, strlen = %d\n", tokens[i], strlen(tokens[i]));
            }
	    //printf("type = %s, ptrlevel = %d\n", typestr, ptrlevel);
            continue;
        }

        vcount++;

        // Determine the variable use
        if (clinetype == CLINE_DECL_STRUCT)
            use   = CVARS_GLOBAL;
        else if (clinetype == CLINE_STRUCT_ELEM)
            use   = CVARS_STRUCT;
        else if (clinetype == CLINE_DECL_LOCAL)
            use   = CVARS_LOCAL;
        else if (clinetype == CLINE_DECL_GLOBAL)
            use   = CVARS_GLOBAL;
        else if (clinetype == CLINE_STRUCT_VAR)
            use   = CVARS_GLOBAL;
        else if (vcount <= 1)
            use   = CVARS_FUNCTION;
        else
            use   = CVARS_PARM;

	// If local or parm check if already on lvarlist
	if ((use == CVARS_LOCAL || use == CVARS_PARM) &&
	    (varinfo = FindVariable(lvarlist, tokens[i])))
	{
	    lvarlist = (VarInfoT *)ListRemoveLink((LinkT *)lvarlist, (LinkT *)varinfo);
            varinfo->signflag = 0;
            varinfo->array = 0;
            varinfo->value = 0;
            varinfo->typeinfo = 0;
	    //printf("Reusing local variable %s (%s)\n", varinfo->str, varinfo->str1);
            //printf("lvarlist\n");
            //ListPrint(lvarlist);
	}
	// else check if function and matches existing prototype
	else if (use == CVARS_FUNCTION && (varinfo = FindVariable(varlist, tokens[i])))
	{
	    if (varinfo->use != CVARS_PROTOTYPE)
	    {
		PrintFileLine();
		printf(" - ERROR: Function name %s has already been used\n", tokens[i]);
		CspinExit(1);
	    }
	    varinfo->use = CVARS_FUNCTION;
	    continue;
	}
	else
	{
            varinfo = (VarInfoT *)LinkCreate(tokens[i], sizeof(VarInfoT));
	    varinfo->str1 = AddToSymbolList(varinfo->str);
	    if (use == CVARS_LOCAL || use == CVARS_PARM)
	    {
	    //printf("Could not find local variable %s on lvarlist\n", varinfo->str);
	    //printf("Created new local variable with alias %s\n", varinfo->str1);
	    }
	}
	varinfo->ptrlevel = ptrlevel;
	//printf("ptrlevel = %d\n", ptrlevel);
	varinfo->use = use;

        // Check variable for brackets
        for (bracketflag = 0; i < num - 1; i++)
        {
            if (bracketflag == 0)
            {
                if (tokens[i+1][0] != '[') break;
                bracketflag = 1;
		firsti = i + 1;
		varinfo->array = 1;
		varinfo->ptrlevel++;
            }
            else if (tokens[i+1][0] == ']')
	    {
		bracketflag = 0;
		temp = EvaluateNumber(&tokens[firsti], i + 2 - firsti);
		if (temp) varinfo->array = temp;
	    }
        }

	if (vartypeinfo == 0)
	{
	    vartypeinfo = FindType(typeslist, "int");
	    if (vartypeinfo == 0)
	    {
		printf("Couldn't find type int in list\n");
		PrintTypes();
		CspinExit(1);
	    }
	}
	varinfo->typeinfo = vartypeinfo;
	varinfo->ptrlevel += vartypeinfo->ptrlevel;

	if (signflag != -1)
	    varinfo->signflag = signflag;
	else
	    varinfo->signflag = vartypeinfo->signflag;

	//printf("ptrlevel = %d\n", varinfo->ptrlevel);
        if (clinetype == CLINE_STRUCT_ELEM)
	{
            TypeInfoT *typeinfo = structlink;
	    if (typeinfo)
	    {
	        int size1 = SizeOfVarInfo(varinfo);
		int size2 = typeinfo->size;
		if (size1 == 2) size2 = (size2 + 1) & 0xfffffffe;
		else if (size1 == 4) size2 = (size2 + 3) & 0xfffffffc;
		varinfo->value = size2;
		typeinfo->size = size1 + size2;
	        typeinfo->structvars = ListAppend(typeinfo->structvars, (LinkT *)varinfo);
		StructElem = varinfo;
		//printf("struct elem %s, %d bytes\n", varinfo->str, size1);
	    }
	    else
	    {
	        PrintFileLine();
	        printf(" - Cannot add elem %s\n", varinfo->str);
	    }
	}
	else
	    varlist = VarInfoAppend(varlist, varinfo);
    }

    return clinetype;
}

void AddIntrinsicFunction(char *type, char *cname, char *sname)
{
    VarInfoT *varinfo = (VarInfoT *)LinkCreate(cname, sizeof(VarInfoT));
    varinfo->str1 = AddToSymbolList(sname);
    varinfo->typeinfo = FindType(typeslist, type);
    varinfo->use = CVARS_FUNCTION;
    varlist = VarInfoAppend(varlist, varinfo);
}

void PrintNewVariableList(FILE *outfile, VarInfoT *varinfo, int lev)
{
    int i, j;
    int ptrlevel;
    TypeInfoT *typeinfo;

    for (i = 0; varinfo; varinfo = varinfo->next, i++)
    {
	typeinfo =  varinfo->typeinfo;
	ptrlevel = varinfo->ptrlevel;
	if (varinfo->array) ptrlevel--;
	for (j = 0; j < lev; j++) fprintf(outfile, "  ");
	if (typeinfo == NULL)
	{
	    fprintf(outfile, "%d: %s | %s\n", i, "BAD", varinfo->str);
	    continue;
	}
	fprintf(outfile, "%d: %s ", i, types[varinfo->use]);
        fprintf(outfile, "%s ", typeinfo->str);
	for(j = 0; j < ptrlevel; j++) fprintf(outfile, "*");
        fprintf(outfile, "%s", varinfo->str);
	if (varinfo->array)
	{
	    fprintf(outfile, "[%d]", varinfo->array);
	}
	if (typeinfo->size)
	    fprintf(outfile, ", size = %d", typeinfo->size);
	fprintf(outfile, "\n");
#ifdef PRINT_STRUCT_ELEM
	if (typeinfo->structvars)
	    PrintNewVariableList(outfile, (VarInfoT *)typeinfo->structvars, lev + 1);
#endif
    }
}

void PrintVariableList(FILE *outfile)
{
    fprintf(outfile, "**** Variable List ****\n");
    PrintNewVariableList(outfile, varlist, 0);
}

// Remove local variables and function parms from varlist and put on lvarlist
void PurgeLocalParms(void)
{
    VarInfoT *next;
    VarInfoT *prev = 0;
    VarInfoT *varinfo = varlist;

//printf("PurgeLocalParms\n");
    while (varinfo)
    {
        //printf("Check %s\n", varinfo->str);
	next = varinfo->next;
	if (varinfo->use == CVARS_LOCAL || varinfo->use == CVARS_PARM)
	{
	    //printf("Add %s to lvarlist\n", varinfo->str);
	    if (prev) prev->next = next;
	    else varlist = next;
	    lvarlist = VarInfoAppend(lvarlist, varinfo);
	    varinfo->next = 0;
    //printf("lvarlist\n");
    //ListPrint(lvarlist);
	}
	else
	    prev = varinfo;

	varinfo = next;
    }
    //printf("lvarlist\n");
    //ListPrint(lvarlist);
}

void PrintTypes(void)
{
    int i = 0;
    TypeInfoT *typeinfo = typeslist;

    printf("Types List\n");
    while (typeinfo)
    {
        printf("%3d: %s, ptrlevel = %d, size = %d, signflag = %d, floatflag = %d", i++,
	typeinfo->str, typeinfo->ptrlevel, typeinfo->size,
	typeinfo->signflag, typeinfo->floatflag);
	if (typeinfo->typedefptr)
	{
	    printf(" - (%s)", typeinfo->typedefptr->str);
	}
	printf("\n");
	typeinfo = typeinfo->next;
    }
}

void AddFuncProtoToVarList(char *type, int ptrlevel, char *name)
{
    VarInfoT *varinfo = (VarInfoT *)LinkCreate(name, sizeof(VarInfoT));
    varinfo->typeinfo = FindType(typeslist, type);
    varinfo->str1 = AddToSymbolList(varinfo->str);
    varinfo->ptrlevel = ptrlevel;
    varinfo->use = CVARS_PROTOTYPE;
    varlist = (VarInfoT *)ListAppend((LinkT *)varlist, (LinkT *)varinfo);
}

void AddCFuncProtoToVarList(char *type, int ptrlevel, char *name)
{
    VarInfoT *varinfo = (VarInfoT *)LinkCreate(name, sizeof(VarInfoT));
    varinfo->typeinfo = FindType(typeslist, type);
    varinfo->str1 = AddToSymbolList(varinfo->str);
    varinfo->ptrlevel = ptrlevel;
    varinfo->use = CVARS_PROTOTYPE;
    varlist = (VarInfoT *)ListAppend((LinkT *)varlist, (LinkT *)varinfo);
    varinfo->str1 = str_alloc(sizeof(name) + 3);
    strcpy(varinfo->str1, "c.");
    strcat(varinfo->str1, name);
}

VarInfoT *AddObjFuncProtoToVarList(char *name)
{
    char *ptr;
    char *type = "int";
    int ptrlevel = 0;
    char objprefix[100];
    VarInfoT *varinfo = (VarInfoT *)LinkCreate(name, sizeof(VarInfoT));
    varinfo->typeinfo = FindType(typeslist, type);
    varinfo->str1 = AddToSymbolList(varinfo->str);
    varinfo->ptrlevel = ptrlevel;
    varinfo->use = CVARS_PROTOTYPE;
    varlist = (VarInfoT *)ListAppend((LinkT *)varlist, (LinkT *)varinfo);
    // Check name for object prefix
    strcpy(objprefix, objname);
    strcat(objprefix, "_");
    if (!objname[0])
    {
        varinfo->str1 = str_alloc(strlen(name) + 1);
        strcpy(varinfo->str1, name);
    }
    else if (!strncmp(name, objprefix, strlen(objprefix)))
    {
        varinfo->str1 = str_alloc(strlen(name) + 1);
        strcpy(varinfo->str1, name);
        ptr = FindChar(varinfo->str1, '_');
        if (*ptr) *ptr = '.';
    }
    else
    {
        varinfo->str1 = str_alloc(strlen(name) + strlen(objprefix) + 1);
        strcpy(varinfo->str1, objname);
        strcat(varinfo->str1, ".");
        strcat(varinfo->str1, name);
    }
    return varinfo;
}
